#pragma once

#if ! defined(WIN32)
typedef int	SOCKET;
typedef struct sockaddr SOCKADDR;
#endif

#include "IGTPoller.h"
#include <arpa/inet.h>
#include <string>
#include <vector>
#include <list>
#include <map>
#include "Mutex.h"
#include "ThreadSyncMngr.h"
#ifdef __USE_ASYNC_HOSTBYNAME__
#include "AsyncHostByName.h"
#endif
#include <sys/socket.h>
#include "IGTEventDelegate.h"
#include "GTLog.h"

#define WM_USER_ASYNC_SOKCET					5000
#define WM_SOCKET_EVENT_ERROR					(WM_USER_ASYNC_SOKCET + 0)
#define WM_SOCKET_EVENT_RECEIVE					(WM_USER_ASYNC_SOKCET + 1)
#define WM_SOCKET_EVENT_SEND					(WM_USER_ASYNC_SOKCET + 2)
#define WM_SOCKET_EVENT_CONNECT					(WM_USER_ASYNC_SOKCET + 3)
#define WM_SOCKET_EVENT_ACCEPT					(WM_USER_ASYNC_SOKCET + 4)
#define WM_SOCKET_EVENT_CLOSE					(WM_USER_ASYNC_SOKCET + 5)


class CAsyncSocketEx 
: public GTPollerClient
, public IGTEventDelegateEvent
, public IGTRunLoopObserver
, public _IThreadSyncObject
#ifdef __USE_ASYNC_HOSTBYNAME__
, public IGTAsyncHostByNameEvent
#endif
{
private:	
	CAsyncSocketEx(const CAsyncSocketEx& rSrc);    // no implementation
	void operator=(const CAsyncSocketEx& rSrc);  // no implementation
	
	// Construction
public:
	CAsyncSocketEx();
	bool Create(unsigned int nSocketPort = 0, int nSocketType = 0/*=SOCK_STREAM*/);
	
	// Attributes
public:
	bool GetPeerName( sockaddr* lpSockAddr, int* lpSockAddrLen );
	bool GetSockName( sockaddr* lpSockAddr, int* lpSockAddrLen );
	

	// Thread Safety
protected:
	virtual void _Retain(void);
	virtual void _Release(void);
	bool _Delete(void);

#define USES_SOCKET_AUTO_RELEASE	KThreadSyncMngr __tmp(m_bDeleteFlag ? NULL : this);
	
	// Operations
public:
	virtual void Delete();

	virtual bool Accept( CAsyncSocketEx& rConnectedSocket, sockaddr* lpSockAddr = NULL, int* lpSockAddrLen = NULL );
	
	bool Bind(unsigned int nSocketPort, const char* lpszSocketAddress = NULL);
	bool Bind(const sockaddr* lpSockAddr, int nSockAddrLen);
	
	bool Listen( int nConnectionBacklog = 5 );
	
	virtual void Close();
		
	virtual bool Connect(const char* lpszHostAddress, unsigned int nHostPort, int nTimeOut = -1 );
	
	virtual int Receive(void* lpBuf, int nBufLen);

	int ReceiveFrom(void* lpBuf, int nBufLen, SOCKADDR* lpSockAddr, socklen_t* lpSockAddrLen, int nFlags = 0);

	virtual int Send(const void* lpBuf, int nBufLen);
	int SendTo(const void* lpBuf, int nBufLen, const sockaddr* lpSockAddr, int nSockAddrLen, int nFlags = 0);
	
	enum { receives = 0, sends = 1, both = 2 };
	bool ShutDown( int nHow = sends );
	
	void Attach(int hSocket);
	int Detach();
	
	void UnCheckOnSendEvent();
	void CheckOnSendEvent();
	void UnCheckOnRecvEvent();
	void CheckOnRecvEvent();
	
	int GetLastError();
	void Shutdown(void);
	static void* ThreadFunc(void *);
	
	int GetSocketHandle(void);

	virtual void SetMainThreadEventMode(bool bEnable);

public:
	// GTPollerClient Callback..
	virtual void OnGTPollerClient_NotifyPollEvent(GTPollerEvent *e);
	virtual void OnGTPollerClient_RegisterEvent(IGTPoller *poller);
	virtual void OnGTPollerClient_UnregisterEvent(IGTPoller *poller);
	
 #ifdef __USE_ASYNC_HOSTBYNAME__
	// IGTAsyncHostByNameEvent Callback..
	virtual void OnGetHostByNameAsyncProgress(const char *szHostName, const char *szIPAddress, void *userinfo) { }
	virtual void OnGetHostByNameAsyncResult(const char *szHostName, int nCount, void *userinfo);
#endif

	// IGTEventDelegateEvent
	virtual void OnGTEventDelegateEvent_DidCreated(IGTEventDelegate *delegate);
	virtual void OnGTEventDelegateEvent_DidReleased(IGTEventDelegate *delegate);
	
	// IGTRunLoopObserver
	virtual bool OnGTRunLoopObserver_SourceEvent(GT_EVENT_HANDLE handle, GT_PARAM param1, GT_PARAM param2);
	virtual bool OnGTRunLoopObserver_TimerEvent(GT_EVENT_HANDLE handle, GT_PARAM param) { }
	
	// Event Post
	void Fire_OnReceive(int nErrorCode, int nBytes);
	void Fire_OnSend(int nErrorCode);
	void Fire_OnOutOfBandData(int nErrorCode);
	void Fire_OnAccept(int nErrorCode);
	void Fire_OnConnect(int nErrorCode);
	void Fire_OnClose(int nErrorCode);
	void Fire_OnSocketError(int nErrorCode);
	
	// Overridable callback & status function
public:
	virtual bool HasSendData() {}
	virtual void OnReceive(int nErrorCode, int nBytes) {}
	virtual void OnSend(int nErrorCode) {}
	virtual void OnOutOfBandData(int nErrorCode) {}
	virtual void OnAccept(int nErrorCode) {}
	virtual void OnConnect(int nErrorCode) {}
	virtual void OnClose(int nErrorCode) {}
	virtual void OnSocketError(int nErrorCode) {}
	
	// Implementation
public:
	virtual ~CAsyncSocketEx();
	
protected:
	friend class _KAutoRelease;
	volatile int m_hSocket;
	int m_nSocketType;

	std::string m_strHostName;
	int m_nCurrentPort;
	int m_nCurrentTimeout;
	volatile bool m_bMainThreadMode;
	
	volatile int32_t m_nRetainCount;	// For Thread Safety
	volatile bool m_bDeleteFlag;		// For Thread Safety
	
	volatile bool m_bSendingFlag;		// For Kqueue Filter Performance
	volatile bool m_bRecvingFlag;		// For Kqueue Filter Performance
	volatile bool m_bFirstPollAction;	// For Nonblocking Connect
	
	volatile bool m_bRequestConnect;	// For Confirm OnConnect()
	volatile bool m_bIsTcpListenSocket;	// For Nonblocking Accept
	
	IGTEventDelegate *m_pEventDelegate;
};


inline void CAsyncSocketEx::_Retain(void)
{
	InterlockedIncrement(&m_nRetainCount);
	
//	GT_PLOG(">>>>>>>>>>>>>>>>>>>>>>>> CAsyncSocketEx::_Retain %p cnt = %d thread %p", this, m_nRetainCount, pthread_self());
}


inline void CAsyncSocketEx::_Release(void)
{
	InterlockedDecrement(&m_nRetainCount);
	
//	GT_PLOG(">>>>>>>>>>>>>>>>>>>>>>>> CAsyncSocketEx::_Release %p cnt = %d thread %p", this, m_nRetainCount, pthread_self());
	
	_Delete();
}


inline bool CAsyncSocketEx::_Delete(void)
{
	if(0 == m_nRetainCount)
	{
		//GT_PLOG("CAsyncSocketEx::_Delete() : 0 == m_nRetainCount %p thread %p delete %d", this, pthread_self(), m_bDeleteFlag);
		
		if(m_bDeleteFlag)
		{
			delete this;				
		}
		else
		{
			//GT_PLOG("CAsyncSocketEx::_Delete() : m_bDeleteFlag = false thread %p", this, pthread_self());
		}
	}
	return false;
}

inline void CAsyncSocketEx::Fire_OnSocketError(int nErrorCode)
{
	if(false == m_bMainThreadMode)
		OnSocketError(nErrorCode);
	else
	{
		// Don't worry about that m_pEventDelegate may be deleted..
		m_pEventDelegate->PostSourceEvent(WM_SOCKET_EVENT_ERROR, (GT_PARAM)nErrorCode, (GT_PARAM)0);
	}
}

inline void CAsyncSocketEx::Fire_OnReceive(int nErrorCode, int nBytes)
{
	if(false == m_bMainThreadMode)
		OnReceive(nErrorCode, nBytes);
	else
	{
		UnCheckOnRecvEvent();
		// Don't worry about that m_pEventDelegate may be deleted..
		m_pEventDelegate->PostSourceEvent(WM_SOCKET_EVENT_RECEIVE, (GT_PARAM)nErrorCode, (GT_PARAM)nBytes);
	}
}


inline void CAsyncSocketEx::Fire_OnSend(int nErrorCode)
{
	if(false == m_bMainThreadMode)
		OnSend(nErrorCode);
	else
	{
		UnCheckOnSendEvent();
		// Don't worry that about m_pEventDelegate may be deleted..
		m_pEventDelegate->PostSourceEvent(WM_SOCKET_EVENT_SEND, (GT_PARAM)nErrorCode, (GT_PARAM)0);
	}	
}


inline void CAsyncSocketEx::Fire_OnAccept(int nErrorCode)
{
	if(false == m_bMainThreadMode)
		OnAccept(nErrorCode);
	else
	{
		// Don't worry about that m_pEventDelegate may be deleted..
		m_pEventDelegate->PostSourceEvent(WM_SOCKET_EVENT_ACCEPT, (GT_PARAM)nErrorCode, (GT_PARAM)0);
	}	
}


inline void CAsyncSocketEx::Fire_OnConnect(int nErrorCode)
{
	if(false == m_bMainThreadMode)
		OnConnect(nErrorCode);
	else
	{
		// Don't worry about that m_pEventDelegate may be deleted..
		m_pEventDelegate->PostSourceEvent(WM_SOCKET_EVENT_CONNECT, (GT_PARAM)nErrorCode, (GT_PARAM)0);
	}	
}


inline void CAsyncSocketEx::Fire_OnClose(int nErrorCode)
{
	if(false == m_bMainThreadMode)
		OnClose(nErrorCode);
	else
	{
		// Don't worry about that m_pEventDelegate may be deleted..
		m_pEventDelegate->PostSourceEvent(WM_SOCKET_EVENT_CLOSE, (GT_PARAM)nErrorCode, (GT_PARAM)0);
	}	
}


inline void CAsyncSocketEx::Fire_OnOutOfBandData(int nErrorCode)
{
	// DEPRECATED..
	ASSERT(false);
}

